import numpy as np
import matplotlib.pyplot as plt
from numba import njit
from scipy import stats

@njit
def energy_spectrum(spectrum: str, k: np.ndarray, urms: float, k0: float) -> np.ndarray:
    """
    Compute the energy spectrum E(k) for a given vector of wavenumber k.

    Parameters
    ----------
    spectrum : string 
        name of the spectrum.
    k : np.ndarray
        Wavenumbers at which to compute the energy spectrum.
    urms : float
        Root mean square velocity.
    k0 : float
        Characteristic wavenumber.

    Returns
    -------
    E_k : float or np.ndarray
        Energy spectrum value(s) at the given wavenumber(s).
    """

    E_k = np.empty_like(k)
    if spectrum == "PassotPouquet":
        E_k = (urms**2 * 16. * np.sqrt(2. / np.pi) * k**4 / k0**5) * np.exp(-2. * (k / k0)**2)
    if spectrum == "Constant":
        E_k = urms*k**4*np.exp(-2.*k**2/k0**2)
    return E_k

def compute_tke_spectrum(u,v,w,lx,ly,lz):
    """
    Compute the turbulent kinetic energy (TKE) spectrum from the velocity fields.

    Parameters
    ----------
    u : np.ndarray
        Velocity component in the x-direction.
    v : np.ndarray
        Velocity component in the y-direction.
    w : np.ndarray
        Velocity component in the z-direction.
    lx : float
        Domain size in the x-direction.
    ly : float
        Domain size in the y-direction.
    lz : float
        Domain size in the z-direction.

    Returns
    -------
    knyquist : float
        Nyquist wavenumber.
    kvals : np.ndarray
        Wavenumbers.
    Abins : np.ndarray
        TKE spectrum values.
    """

    # --- Size
    n0, n1, n2 = u.shape
    L           = [lx, ly, lz]
    oneOverN = 1./(n0*n1*n2)
		
    # --- FFT
    fourier_u          = np.fft.fftn(u)*oneOverN
    fourier_v          = np.fft.fftn(v)*oneOverN
    fourier_w          = np.fft.fftn(w)*oneOverN
    fourier_amplitudes = 0.5*(np.abs(fourier_u)**2 + np.abs(fourier_v)**2 + np.abs(fourier_w)**2)
        
    # --- Frequency
    kfreq0             = np.fft.fftfreq(n0) * n0 * (2*np.pi/L[0])
    kfreq1             = np.fft.fftfreq(n1) * n1 * (2*np.pi/L[1])
    kfreq2             = np.fft.fftfreq(n2) * n2 * (2*np.pi/L[2])
    kfreq3D            = np.meshgrid(kfreq0, kfreq1, kfreq2)
    knrm               = np.sqrt(kfreq3D[0]**2 + kfreq3D[1]**2 + kfreq3D[2]**2)

    # --- Flatten vector
    knrm               = knrm.flatten()
    fourier_amplitudes = fourier_amplitudes.flatten()
    kbins              = np.arange(0, min(n0,n1,n2)//2+1, 1.)*(2*np.pi/np.min(np.array(L)))
    kvals              = kbins[:-1] 

    # --- Spectrum 
    Abins, _, _        = stats.binned_statistic(knrm, fourier_amplitudes, statistic = "sum", bins = kbins)

    knyquist = knrm * min(n0, n1, n2) / 2
	
    return knyquist, kvals, Abins/(kbins[1:] - kbins[:-1])

def plot_spectrum(wmax, wave_numbers, tke_spectrum, real_spectrum=None, save_path=None):
    plt.figure()
    plt.loglog(wave_numbers, tke_spectrum, 'o-', label='Computed Spectrum')
    plt.axvline(x=wmax, color='r', linestyle='--', label='Cut-off Wavenumber')
    
    if real_spectrum is not None:
        plt.loglog(wave_numbers, real_spectrum, 'k', label='Real Spectrum')
    
    plt.xlabel('Wavenumber (k)')
    plt.ylabel('Energy Spectrum (E(k))')
    plt.ylim(np.min(tke_spectrum), 1.5*np.max(tke_spectrum))
    plt.legend()
    plt.grid()
    if save_path:
        plt.savefig(save_path)
    else:
        plt.show()
    plt.close()
